#include <GLUT/glut.h>
#include <math.h>
#include <iostream>
#include <map>
#include <tr1/unordered_set>
#include <list>
#include "UnstructuredGrid3D.h"
#include "TensorVolume.h"
#include "MeshRenderer.h"
#include "TensorGlyphRenderer.h"
#include "Isosurface.h"
#include "pca.h"
#include "Point3d.h"
#include "zpr.h"

using namespace std;




TensorVolume<float>  tensor;							//Tensor field to visualize
Volume<float>		 fa;								//Fractional anisotropy (Basser and Pierpaoli '96)
Volume<Point3d>*	 eigenvectors[3];
	
UnstructuredGrid3D*	 isosurface		 = 0;				//2D mesh storing the isosurface of the tensor's confidence field
Isosurface<float>	 isoEngine;							//Engine that extracts an isosurface (as an UnstructuredGrid3D) from a 3D scalar volume (stored as a Volume<T>)
MeshRenderer		 isoRenderer;						//Renderer to draw the extracted isosurface as a shaded mesh
TensorGlyphRenderer  glyphRenderer;						//Renderer to draw the tensor glyphs

float				 max_depth		 = 0;
bool				 show_isosurface = true;			//Whether to show the confidence isosurface or not
int					 max_glyphs		 = 2000;			//Maximum number of glyphs we want to draw
int					 subsampling	 = 3;				//Minimal distance (in voxels) between two glyph-seeds. Used to avoid a too high
														//crowding of the drawn glyphs.



static void mouseclick(int button, int state, int x, int y)
{
	int keys = glutGetModifiers();						//The mouse events are interpreted as follows:	
	if (keys & GLUT_ACTIVE_CTRL)						// 
	button = GLUT_MIDDLE_BUTTON;						// -left button + move:                         rotate viewpoint
	if (keys & GLUT_ACTIVE_SHIFT)						// -middle button (or left button+Control key): zoom in / out
	button = GLUT_RIGHT_BUTTON;							// -right buttom (or left button+Shift key):    translate viewpoint

	zprMouse(button,state,x,y);							//Use the ZPR library to manipulate the viewpoint. The library sets the modelview  
														//OpenGL matrix based on the mouse events, thereby changing the current viewpoint.
}



static void mousemotion(int x, int y)
{
	zprMotion(x,y);
	glutPostRedisplay();
}


static void viewing(int width, int height)					//GLUT window-reshape callback
{
	glutReshapeWindow(width,height);  
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0,double(width)/height,10.0f,max_depth);
	glViewport(0,0,width,height);
}


static void display()										//GLUT display callback
{
   glClearColor(1,1,1,1);									//1. Clear the frame and depth buffers
   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   const int width = tensor.getWidth(), height = tensor.getHeight(), depth = tensor.getDepth();

   glMatrixMode(GL_MODELVIEW);	
   glPushMatrix();
   glTranslatef(-width/2,-height/2,-1.7*depth);				//2. Translate the volume so it's centered and in front of the near clipping plane

   if (show_isosurface)										//3. Show the isosurface (half-transparent)
      isoRenderer.draw(*isosurface);						
   glyphRenderer.draw(eigenvectors);						//4. Show the tensor glyphs
															
   glPopMatrix();   
   glutSwapBuffers();
}




void keyboard(unsigned char c,int,int)						//Press space to cycle through the renderer's drawing styles and colormaps
{
	switch(c)
	{
	case ' ':												//Show or hide the confidence isosurface
		{
		  show_isosurface = !show_isosurface;
		  break;
		}  	
	case '+':												//Draw more glyphs
	case '=':
	    {
		  glyphRenderer.setMaxSeeds(glyphRenderer.getMaxSeeds()+1);
		  break;
		}  
	case '-':												//Draw less glyphs
	case '_':
	    {
		  glyphRenderer.setMaxSeeds(glyphRenderer.getMaxSeeds()-1);
		  break;
		}  
	}
	
	glutPostRedisplay();    
}	




void computePCA(TensorVolume<float>& f, int i, int j, int k, Point3d& e0, Point3d& e1, Point3d& e2, float& l0, float& l1, float& l2)
{																	//Given a 3D tensor volume, compute its eigenvectors and eigenvalues
																	//at each voxel.
    float *M[3],*V[3],W[3],MM0[3],MM1[3],MM2[3],VV0[3],VV1[3],VV2[3]; 
  
    M[0] = MM0; M[1] = MM1; M[2] = MM2; V[0] = VV0; V[1] = VV1; V[2] = VV2; 

																	//Fill-in symmetric M[][] for eigen-computation
    M[0][0] = f.dxx(i,j,k);  M[0][1] = f.dxy(i,j,k);  M[0][2] = f.dxz(i,j,k);	
    M[1][0] = M[0][1];       M[1][1] = f.dyy(i,j,k);  M[1][2] = f.dyz(i,j,k);
    M[2][0] = M[0][2];		 M[2][1] = M[1][2];		  M[2][2] = f.dzz(i,j,k);
    
    JacobiN(M,3,W,V);												//Compute the eigenvalues W[] and vectors V[] of M[][]
																	//Results are returned in decreasing eigenval-order W[0]>W[1]>W[2]
    W[0] = fabs(W[0]); W[1] = fabs(W[1]); W[2] = fabs(W[2]);
																	//Return eigenvector and eigenvalue results
	e0   = Point3d(VV0[0],VV1[0],VV2[0]); e0.normalize(); e0 *= W[0];
	e1   = Point3d(VV0[1],VV1[1],VV2[1]); e1.normalize(); e1 *= W[1];
	e2   = Point3d(VV0[2],VV1[2],VV2[2]); e2.normalize(); e2 *= W[2];
		
	l0 = W[0];   
	l1 = W[1];	
	l2 = W[2];	
}




void computeMetrics(TensorVolume<float>& tensor,Volume<float>& fa,Volume<Point3d>* eigenvectors[])
{														//Compute various tensor-field metrics, and save them for later use
	Point3d E0,E1,E2; float l0,l1,l2;
	
	int D=tensor.getDepth(),H=tensor.getHeight(),W=tensor.getWidth(); 
	
	fa.makeVolume(W,H,D);								//Allocate volumes for the tensor metrics we will compute
	eigenvectors[0] = new Volume<Point3d>(W,H,D);
	eigenvectors[1] = new Volume<Point3d>(W,H,D);
	eigenvectors[2] = new Volume<Point3d>(W,H,D);

	for(int k=0;k<D;++k)								//Compute metrics. Mask all of them with the tensor's confidence,	
	for(int j=0;j<H;++j)								//since we don't care to visualize anything where the confidence is low.
	for(int i=0;i<W;++i)	
	{
	   computePCA(tensor,i,j,k,E0,E1,E2,l0,l1,l2);
	   
	   float  c = tensor.confidence(i,j,k); 
	   float md = (l0+l1+l2)/3;							
	   float FA = sqrt(1.5*((l0-md)*(l0-md)+(l1-md)*(l1-md)+(l2-md)*(l2-md))/(l0*l0+l1*l1+l2*l2));
															
	   fa(i,j,k)					= c*FA;				//Fractional anisotropy metric
	   (*eigenvectors[0])(i,j,k)	= E0*c;				//Major eigenvector metric	   
	   (*eigenvectors[1])(i,j,k)    = E1*c;				//Medium eigenvector metric	   
	   (*eigenvectors[2])(i,j,k)    = E2*c;				//Minor eigenvector metric	   
	}	
	
	float fa_min,fa_max;								//Normalize computed metrics
	fa.getMinMax(fa_min,fa_max);						//(so we can use them easier next for visualization)

	for(int k=0;k<D;++k)	
	for(int j=0;j<H;++j)	
	for(int i=0;i<W;++i)	
		fa(i,j,k)		   = (fa(i,j,k)-fa_min)/(fa_max-fa_min);
}




void computeSeeds(Volume<float>& confidence, Volume<float>& anisotropy,int maxGlyphs,int subSample)
{														//Compute 'maxGlyphs' seeds in the tensor field for placing tensor glyphs.
														//Seeds are computed in regions of maximum confidence and anisotropy.

	tr1::unordered_set<unsigned int> visited;			//Keeps track of where (in the volume) glyph-seeds have been placed. When we
														//place a seed at location (i,j,k), we add a unique code for (i,j,k) to the
														//'visited' set. When we try to place a new glyph (a,b,c), we only do this if
														//'visited' does not contain the code for (a,b,c).

	int W = confidence.getWidth(), H = confidence.getHeight(), D = confidence.getDepth();	
	multimap<float,coord3> potentialSeeds;				//Stores candidate seed-points, sorted by anisotropy

	for(int k=0;k<D;++k)								//1. Compute candidate seed-points. These are all volume points which
	for(int j=0;j<H;++j)								//a) have a confidence larger than 0.5
	for(int i=0;i<W;++i)								//b) have an anisotropy larger than 0.5
	{													//   This way, we limit the number of computed candidate seed-points AND also
														//	 consider the most likely ones to show interesting tensor values
		int I = i-i%subSample, 
		    J = j-j%subSample,							//2. Separate seed-points by a minimal distance 'subSample'. 
			K = k-k%subSample;							//   For this, we simply create them on a uniformly subsampled version of
														//   the initial grid. 

		if (confidence(I,J,K)<0.5) continue;			
		float key = anisotropy(I,J,K);
		if (key<0.5) continue;
		
		unsigned int code = I+(unsigned int)W*J+(unsigned int)W*H*K;
		if (visited.find(code)!=visited.end()) continue;//3. Do not place two seeds at the same (subsampled) position, since we
														//   can only draw a single glyph there
		
		potentialSeeds.insert(make_pair(key,coord3(I,J,K)));
		visited.insert(code);							//4. Add a new potential seed
	}
	
	vector<Point3d>	seeds;								//5. From the candidate seed-points, select the 'maxGlyphs' ones having
														//   the largest anisotropy values. 
	for(multimap<float,coord3>::reverse_iterator it=potentialSeeds.rbegin();it!=potentialSeeds.rend() && maxGlyphs>0;--maxGlyphs,++it)
	{
		const coord3& c = it->second;
		seeds.push_back(Point3d(c.x,c.y,c.z));
	}
	
	glyphRenderer.setSeeds(seeds);						//6. Pass seeds to the glyph renderer	
	glyphRenderer.setMaxSeeds(seeds.size());			//   Also set other parameters of the glyph renderer	
	glyphRenderer.setScale(1.5);
}




int main(int argc,char* argv[])								//Main program
{
	cout<<"6. Tensor glyphs"<<endl;
	cout<<"In the graphics window, press"<<endl;
	cout<<"   space: toggle the isosurface of the tensor confidence on/off"<<endl;
	cout<<"   +,-:   show more or less glyphs"<<endl;


    glutInit(&argc, argv);									//1. Initialize the GLUT toolkit
    glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE|GLUT_DEPTH);	//2. Ask GLUT to create next windows with a RGB framebuffer and a Z-buffer too
    glutInitWindowSize(500,500);							//3. Tell GLUT how large are the windows we want to create next
    glutCreateWindow("6. Tensor glyphs");
	
    const char* input_file = (argc>1)? argv[1] : "DATA/gk2-rcc-mask.nhdr";
	
    bool ok = tensor.readNRRD(input_file);					//4. Read the tensor field from an NRRD-format file														
	if (!ok)
	{
		cout<<"Error reading tensor volume "<<input_file<<", aborting"<<endl;
		exit(1);
	}	

	int volSize = (tensor.getWidth()+tensor.getHeight()+tensor.getDepth())/3.0;

	computeMetrics(tensor,fa,eigenvectors);					//5. Compute tensor metrics
	
	isosurface = isoEngine.Contour(tensor.confidence,0.5);	//6. Construct and save confidence isosurface into 'grid'
	isosurface->computeFaceNormals();							
	isosurface->computeVertexNormals();						
	isoRenderer.setDrawingStyle(MeshRenderer::DRAW_C1_CELLS);

    computeSeeds(tensor.confidence,fa,max_glyphs,subsampling);
															//7. Compute locations (seeds) for the tensor glyphs

	zprInit(0,0,-(1.7f-0.5f)*volSize);
	max_depth = 3.0*volSize;

	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouseclick);
	glutMotionFunc(mousemotion);
	glutReshapeFunc(viewing);
	glutDisplayFunc(display);
    glutMainLoop();											

	return 0;
}




